import pathlib

NPEXP_ROOT = pathlib.Path("//allen/programs/mindscope/workgroups/np-exp")
INCOMING_ROOT = pathlib.Path(
    "//allen/programs/braintv/production/incoming/neuralcoding"
)

TEMPLATES_ROOT = pathlib.Path(
    "//allen/programs/mindscope/workgroups/dynamicrouting/ben/npexp_data_manifests"
)

MVR_RELATIVE_PATH = pathlib.Path("c$/ProgramData/AIBS_MPE/mvr/data")
NEWSCALE_RELATIVE_PATH = pathlib.Path("c$/MPM_data")
CAMVIEWER_RELATIVE_PATH = pathlib.Path("c$/Users/svc_neuropix/cv3dImages")  # NP.0 only
CAMSTIM_RELATIVE_PATH = pathlib.Path("c$/ProgramData/AIBS_MPE/camstim/data")
SYNC_RELATIVE_PATH = pathlib.Path("c$/ProgramData/AIBS_MPE/sync/data")

NEUROPIXELS_DATA_RELATIVE_PATH = pathlib.Path(
    "c$/ProgramData/AIBS_MPE/neuropixels_data"
)
NPEXP_PATH = pathlib.Path("//allen/programs/mindscope/workgroups/np-exp")

QC_PATHS = (
    pathlib.Path(
        "//allen/programs/braintv/workgroups/nc-ophys/corbettb/NP_behavior_pipeline/QC"
    ),
    pathlib.Path("//allen/programs/mindscope/workgroups/openscope/GLO_QC"),
    pathlib.Path("//allen/programs/mindscope/workgroups/openscope/Illusion_QC"),
)
"Item 0 is used as default - currently Corbett's folder."



from __future__ import annotations

import datetime
import json
import warnings
from typing import Optional, Union

import requests


class MouseNotInMTrainError(Exception):
    pass


class MTrain:

    server = "http://mtrain:5000"

    @classmethod
    def connected(cls) -> bool:
        response = requests.get(cls.server)
        return True if response.status_code == 200 else False

    def __init__(self, mouse_id: Optional[int | str] = None):
        # we'll only allow mouse_id to be set once per instance
        # but it doesn't necessarily have to be set on init
        if mouse_id:
            self.mouse_id = mouse_id
    
    def __repr__(self):
        return f"{self.__class__.__name__}({self.mouse_id!r})"
    
    def session(self):
        session = requests.session()
        session.post(
            self.server,
            data={
                "username": "chrism",
                "password": "password",
            },
        )
        return session

    @property
    def mouse_id(self) -> str:
        return self._mouse_id

    @mouse_id.setter
    def mouse_id(self, value: int | str):
        if hasattr(self, "_mouse_id") and self._mouse_id is not None:
            raise ValueError("Mouse ID can only be set once per instance.")
        response = requests.get(
            f"{self.server}/get_script/", data=json.dumps({"LabTracks_ID": str(value)})
        )
        if response.status_code == 200:
            self._mouse_id = str(value)
        else:
            raise MouseNotInMTrainError(f"Mouse ID {value} not found in MTrain")

    @property
    def state(self) -> dict[str, int]:
        """Returns dict with values 'id', 'regimen_id', 'stage_id' - all ints"""
        return requests.get(f"{self.server}/api/v1/subjects/{self.mouse_id}").json()[
            "state"
        ]

    @state.setter
    def state(self, value: Union[dict, int, str]):
        """Allows switching to one of the states in the current regimen.

        To change regimen, use set_regimen_and_stage()

        Requires a state dict, state id (str/int) or stage name (str)
        """
        valid_dict = valid_id = None

        if isinstance(value, dict):
            if value in self.states:
                valid_dict = value
            elif (
                hasattr(value, "regimen_id")
                and (str(value["regimen_id"]) in self.all_regimens().keys())
                and not (value["regimen_id"] == self.regimen["id"])
            ):
                warnings.warn(
                    "Trying to change regimen via set state method: use 'obj.regimen=...' instead"
                )
                return

        elif isinstance(value, int) or value.isdigit():
            # check if input matches a state id
            valid_id = (
                int(value)
                if str(value) in [str(s["id"]) for s in self.states]
                else None
            )

        elif isinstance(value, str) and value.lower() in [
            str(s["name"]).lower() for s in self.stages
        ]:
            # check if input matches a stage name
            valid_id = [s["id"] for s in self.stages if s["name"] == value][0]

        if not any([valid_dict, valid_id]):
            warnings.warn(
                f"No matching state found in regimen {self.regimen['name']}. Check obj.states"
            )
            return

        if valid_id and not valid_dict:
            value = [s for s in self.states if s["id"] == value][0]

        with self.session() as s:
            s.post(
                f"{self.server}/set_state/{self.mouse_id}",
                data={
                    "state": json.dumps(value),
                },
            )
        assert self.state == value, "set state failed!"

    @property
    def states(self):
        return self.regimen["states"]

    @property
    def regimen(self) -> dict:
        """Returns dictionary containing 'id', 'name', 'stages', 'states'"""
        return requests.get(
            f"{self.server}/api/v1/regimens/{self.state['regimen_id']}"
        ).json()

    @property
    def all_behavior_sessions(self) -> dict:
        """Returns dictionary containing details of all behavior sessions for mouse_id"""
        result = requests.get(
            f"{self.server}/get_behavior_sessions/",
            data=json.dumps({"LabTracks_ID": str(self.mouse_id)}),
        )

        if result:
            return result.json()["results"]
        return None

    def last_behavior_session_on(self, query_date: datetime.date) -> dict:
        all_behavior_sessions = self.all_behavior_sessions
        matching_sessions = []
        for session in all_behavior_sessions:
            session_datetime = datetime.datetime.strptime(
                session["date"], "%a, %d %b %Y %H:%M:%S %Z"
            )
            if session_datetime.date() == query_date:
                matching_sessions.append(session)

        if matching_sessions:
            return matching_sessions[-1]
        return None

    @regimen.setter
    def regimen(self):
        warnings.warn(
            "Setting the regimen is disabled: use 'obj.set_regimen_and_stage()' to explicitly change both together"
        )

    def set_regimen_and_stage(
        self, regimen: Union[dict, int, str] = None, stage: Union[dict, int, str] = None
    ):
        """Requires a regimen dict, id (str/int) or name (str), plus stage dict, id (str/int) or name (str)"""

        def warn_and_return():
            warnings.warn(
                "Regimen not changed: invalid input. Provide an identifier for both a regimen and a stage."
            )
            return

        if not regimen and stage:
            warn_and_return()

        # we're not setting the regimen directly - we just need some info to set a valid state
        regimen_id: int = None
        stage_id: int = None

        if isinstance(regimen, dict):
            if regimen in self.get_all("regimens"):
                # valid dict provided
                regimen_id = regimen["id"]
            else:
                warn_and_return()

        elif str(regimen) in self.all_regimens().keys():
            # valid id provided
            regimen_id = int(regimen)

        elif any(str(regimen) == s.lower() for s in self.all_regimens().values()):
            # valid name provided
            regimen_id = [
                s["id"]
                for s in self.get_all("regimens")
                if s["name"].lower() == str(regimen).lower()
            ][0]

        else:
            warn_and_return()

        # get the stages available in the new regimen, without setting anything yet
        # new_regimen = self.get_all("regimens")['id'==regimen_id]
        new_regimen = [s for s in self.get_all("regimens") if s["id"] == regimen_id][0]
        new_stages = new_regimen["stages"]

        if isinstance(stage, dict):
            if stage in new_stages:
                stage_id = stage["id"]  # valid dict provided
            else:
                warn_and_return()

        elif str(stage) in [str(s["id"]) for s in new_stages]:
            stage_id = int(stage)  # valid id provided

        elif str(stage).lower() in [str(s["name"]).lower() for s in new_stages]:
            stage_id = [s["id"] for s in new_stages if s["name"] == stage][
                0
            ]  # valid name provided

        else:
            warn_and_return()

        # look for corresponding state
        state_match = [
            s
            for s in new_regimen["states"]
            if s["regimen_id"] == regimen_id and s["stage_id"] == stage_id
        ]
        if not state_match or len(state_match) != 1:
            warn_and_return()
        new_state = state_match[0]

        # set directly instead of using state set method (which intentionally blocks setting a state dict with a new regimen)
        with self.session() as s:
            s.post(
                f"{self.server}/set_state/{self.mouse_id}",
                data={
                    "state": json.dumps(new_state),
                },
            )
        assert self.state == new_state, "set regimen and stage failed!"

    @property
    def stage(self):
        for item in self.stages:
            if item["id"] == self.state["stage_id"]:
                return item

    @stage.setter
    def stage(self, value: Union[str, int]):
        """Accepts stage id (int/str) or name (str)"""

        state_match = None

        if isinstance(value, int) or value.isdigit():
            value = int(value)
            state_match = [s for s in self.states if s["stage_id"] == value]

        if isinstance(value, str):
            stage_match = [s for s in self.stages if s["name"].lower() == value.lower()]
            state_match = [
                s for s in self.states if s["stage_id"] == stage_match[0]["id"]
            ]

        if not state_match or len(state_match) != 1:
            warnings.warn(
                f'No state with stage name or id {value} found in regimen {self.regimen["name"]}. Check obj.stages or obj.states'
            )
            return

        self.state = state_match[0]

    @property
    def script(self) -> dict:
        """Re-routes to stage property"""
        return self.stage

    @script.setter
    def script(self, value: Union[str, int]) -> dict:
        """Accepts stage id (int/str) or name (str)
        Re-routes to stage property
        """
        self.stage = value

    @property
    def stages(self):
        return self.regimen["stages"]

    @classmethod
    def paginated_get(cls, route, page_size=10, offset=0):
        page_number = offset + 1  # page number is 1 based, offset is page
        result = requests.get(
            f"{route}?results_per_page={page_size}&page={page_number}"
        ).json()
        total_pages = result["total_pages"]
        if page_number == total_pages:
            new_offset = None
            has_more = False
        else:
            new_offset = offset + 1
            has_more = True

        return result["objects"], has_more, new_offset

    @classmethod
    def get_all(cls, endpoint: str):
        # page_size = 200 is over the actual page size limit for the api but we don't appear to know what that value is
        # ? 'total_pages': 18
        if endpoint not in ["regimens", "states", "stages", "subjects"]:
            raise ValueError(f"Endpoint {endpoint} not recognized")
        all = []
        max_fetch = 100
        page_size = 200
        offset = 0
        for _ in range(max_fetch):
            results, has_more, new_offset = cls.paginated_get(
                f"{cls.server}/api/v1/{endpoint}", page_size, offset
            )
            all.extend(results)
            if not has_more:
                break
            offset = new_offset
        else:
            warnings.warn(f"Failed to get full list of {endpoint}.")
        return all

    @classmethod
    def all_regimens(cls) -> dict:
        """List of dicts {str(regimen['id']):regimen['name']}"""
        d = {}
        for val in cls.get_all("regimens"):
            d.update({str(val["id"]): val["name"]})
        return d

    # the two methods below were added after the others above:
    # other functions could be re-written to use them for cleaner code

    @classmethod
    def all_states(cls) -> dict:
        """dict containing {str(state['id']):{'id':int,'regimen_id':int,'stage_id':int}}"""
        d = {}
        for val in cls.get_all("states"):
            d.update({str(val["id"]): val})
        return d

    @classmethod
    def all_stages(cls) -> dict:
        """Takes a while. dict containing {str(stage['id']):stage['name']}"""
        d = {}
        for val in cls.get_all("stages"):
            d.update({str(val["id"]): val["name"]})
        return d


if __name__ == "__main__":
    # print(MTrain.connected())
    x = MTrain("632296")
    x.last_behavior_session_on(query_date=datetime.date(2022, 9, 27))
    x.state = x.regimen["states"][0]
    x.state = 1734
    x.stage = 1751
    illusion_regimens = [r for r in x.get_all("regimens") if "Illusion" in r["name"]]
    # x.regimen = x.get_all("regimens")[4]
    x.all_stages()
    x.all_states()
    # x.set_regimen_and_stage(illusion_regimens[4], illusion_regimens[4]['stages'][4])



from __future__ import annotations

import abc
import collections
import datetime
import enum
import functools
import json
import pathlib
import re
from typing import Any, Callable, ClassVar, Type, Union

from np_session.databases.lims2 import (LIMS2MouseInfo, LIMS2ProjectInfo,
                                        LIMS2UserInfo)
from np_session.databases.mtrain import MTrain


class InfoBaseClass(abc.ABC):
    "Store details for an object from various databases. The commonly-used format of its name, e.g. '366122' for a mouse ID, can be obtained by converting to str()."

    id: int | str
    "Commonly-used format of the object's value among the neuropixels team e.g. for a mouse -> the labtracks ID (366122)."
    def __str__(self) -> str:
        return str(self.id)
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id!r})"
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return self.id == str(other)
        if isinstance(other, int):
            return self.id == int(str(other))
        return False

class Mouse(InfoBaseClass):
    def __init__(self, labtracks_mouse_id: str | int):
        self.id = int(labtracks_mouse_id)
    
    @property
    def lims(self) -> LIMS2MouseInfo | dict:
        "Lims info for the mouse."
        if not hasattr(self, "_lims"):
            try:
                self._lims = LIMS2MouseInfo(self.id)
            except ValueError:
                self._lims = {}
        return self._lims
    
    @functools.cached_property
    def mtrain(self) -> MTrain:
        "Lims info for the mouse."
        return MTrain(self.id)
    
    
class User(InfoBaseClass):
    def __init__(self, lims_user_id: str):
        self.id = str(lims_user_id)
        
    @functools.cached_property
    def lims(self) -> LIMS2UserInfo | dict:
        "Lims info for the user."
        if not hasattr(self, "_lims"):
            try:
                self._lims = LIMS2UserInfo(self.id)
            except ValueError:
                self._lims = {}
        return self._lims

class Projects(enum.Enum):
    "All specific project names (used on lims) associated with each umbrella project."
    
    VAR = (
        "VariabilitySpontaneous", 
        "VariabilityAim1",
        )
    GLO = (
        "OpenScopeGlobalLocalOddball",
        )
    ILLUSION = (
        "OpenScopeIllusion",
        )
    DR = (
        "DynamicRoutingSurgicalDevelopment",
        "DynamicRoutingDynamicGating",
        "DynamicRoutingTask1Production",
    )
    VB = (
        'NeuropixelVisualBehavior',
    )
    
class Project(InfoBaseClass):
    
    def __init__(self, lims_project_name: str):
        self.id = str(lims_project_name)
        
    @functools.cached_property
    def lims(self) -> LIMS2ProjectInfo:
        "Lims info for the project."
        return LIMS2ProjectInfo(self.id)
    
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Projects):
            return self.id in other.value
        return super().__eq__(other)
    


from __future__ import annotations

import contextlib
import datetime
import doctest
import functools
import os
import pathlib
from typing import Any, Generator, Union

import np_config
import np_logging
from typing_extensions import Literal

from np_session.components.info import Mouse, Project, Projects, User
from np_session.components.paths import *
from np_session.databases import data_getters as dg
from np_session.databases import lims2 as lims
from np_session.databases import mtrain
from np_session.utils import *

logger = np_logging.getLogger(__name__)

PathLike = Union[str, bytes, os.PathLike, pathlib.Path]
# https://peps.python.org/pep-0519/#provide-specific-type-hinting-support
# PathLike inputs are converted to pathlib.Path objects for os-agnostic filesystem operations.
# os.fsdecode(path: PathLike) is used where only a string is required.


class SessionError(ValueError):
    """Raised when a session folder string ([lims-id]_[mouse-id]_[date]) can't be found in a
    filepath"""

    pass


class FilepathIsDirError(ValueError):
    """Raised when a directory is specified but a filepath is required"""

    pass


class Session:
    """Session information from any string or PathLike containing a lims session ID.

    Note: lims/mtrain properties may be empty or None if mouse/session isn't in db.

    Quick access to useful properties:
    >>> session = Session('c:/1116941914_surface-image1-left.png')
    >>> session.lims.id
    1116941914
    >>> session.folder
    '1116941914_576323_20210721'
    >>> session.project.lims.id
    714916854
    >>> session.is_ecephys_session
    True
    >>> session.rig.acq # hostnames reflect the computers used during the session, not necessarily the current machines
    'W10DT05515'

    Some properties are returned as objects with richer information:
    - `pathlib` objects for filesystem paths:
    >>> session.lims_path.as_posix()
    '//allen/programs/braintv/production/visualbehavior/prod0/specimen_1098595957/ecephys_session_1116941914'
    >>> session.data_dict['storage_directory'].as_posix()
    '//allen/programs/braintv/production/visualbehavior/prod0/specimen_1098595957/ecephys_session_1116941914'


    - `datetime` objects for easy date manipulation:
    >>> session.date
    datetime.date(2021, 7, 21)
    
    - dictionaries from lims (loaded lazily):
    >>> session.mouse
    Mouse(576323)
    >>> session.mouse.lims
    LIMS2MouseInfo(576323)
    >>> session.mouse.lims.id
    1098595957
    >>> session.mouse.lims['full_genotype']
    'wt/wt'

    ...with a useful string representation:
    >>> str(session.mouse)
    '576323'
    >>> str(session.project)
    'NeuropixelVisualBehavior'
    >>> str(session.rig)        # see np_config.Rig
    'NP.0'
    """
    def __lt__(self, other: Session) -> bool:
        if not hasattr(other, 'date'):
            return NotImplemented
        return self.date < other.date

    def __str__(self) -> str:
        return self.folder
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.folder!r})'

    def __init__(self, path_or_session: PathLike | int | LIMS2SessionInfo):
        
        path_or_session = str(path_or_session)
        
        path_or_session = pathlib.Path(path_or_session)
        
        np_folder = folder(path_or_session)

        if not np_folder:
            np_folder = folder_from_lims_id(path_or_session)

        if np_folder is None:
            raise SessionError(f"{path_or_session} does not contain a valid lims session id or session folder string")

        self.folder = np_folder
        self.id = int(self.folder.split("_")[0])
        
        if isinstance(path_or_session, LIMS2SessionInfo):
            self._lims = path_or_session

    @property
    def lims(self) -> lims.LIMS2SessionInfo | dict:
        """
        >>> info = Session(1116941914).lims
        >>> info['stimulus_name']
        'EPHYS_1_images_H_3uL_reward'
        >>> info['operator']['login']
        'taminar'

        >>> Session(1116941914).lims
        LIMS2SessionInfo(1116941914)
        >>> str(Session(1116941914).lims)
        '1116941914'
        """
        if not hasattr(self, "_lims"):
            try:
                self._lims = lims.LIMS2SessionInfo(self.id)
            except ValueError:
                self._lims = {}
        return self._lims

    @property
    def mouse(self) -> Mouse:
        if not hasattr(self, "_mouse"):
            self._mouse = Mouse(self.folder.split("_")[1])
        return self._mouse
    
    @property
    def user(self) -> User | None:
        if not hasattr(self, "_user"):
            lims_user_id = self.lims.get('operator', {}).get('login', '')
            if lims_user_id:
                self._user = User(lims_user_id)
            else:
                self._user = None
        return self._user
    
    @functools.cached_property
    def date(self) -> datetime.date:
        d = self.folder.split("_")[2]
        date = datetime.date(year=int(d[:4]), month=int(d[4:6]), day=int(d[6:]))
        return date

    @property
    def rig(self) -> np_config.Rig | None:
        "Rig object with computer info and paths, can also be used as a string."
        if not hasattr(self, "_rig"):
            self._rig = None
            while not self.rig:
                
                # try from current rig first
                with contextlib.suppress(ValueError):
                    self.rig = np_config.Rig()
                    continue
                    
                # TODO try from platform json
                
                # try from lims 
                rig_id: str | None = self.data_dict.get('rig')
                if rig_id:
                    self.rig = np_config.Rig(rig_id)
                    continue
                
                break
        return self._rig
    
    @rig.setter
    def rig(self, value: np_config.Rig) -> None:
        if not isinstance(value, np_config.Rig):
            raise TypeError(f'Expected `rig` to be an instance of `np_config.Rig`, not {type(value)}')
        self._rig = value
        self.update_hostnames_for_replaced_computers()

    def update_hostnames_for_replaced_computers(self) -> None:
        for comp in ("sync", "stim", "mon", "acq"):
            if replaced := old_hostname(f"{self._rig.id}-{comp.capitalize()}", self.date):
                setattr(self._rig, f"_{comp}", replaced)

    @property
    def is_ecephys_session(self) -> bool | None:
        """False if behavior session, None if unsure."""
        if not self.lims:
            return None
        return "ecephys_session" in self.lims.get("storage_directory", "")

    @property
    def npexp_path(self) -> pathlib.Path:
        """np-exp root / folder (may not exist)"""
        return NPEXP_ROOT / self.folder

    @property
    def lims_path(self) -> pathlib.Path | None:
        """Corresponding directory in lims, if one can be found"""
        if not hasattr(self, "_lims_path"):
            path: str = self.lims.get("storage_directory", "")
            if not path:
                logger.debug(
                    "lims checked successfully, but no folder uploaded for ", self.id
                )
                self._lims_path = None
            else:
                self._lims_path = pathlib.Path("/" + path)
        return self._lims_path
    
    @property
    def z_path(self) -> pathlib.Path:
        "Path in Sync neuropixels_data (aka Z:) (may not exist)) "
        return np_config.local_to_unc(self.rig.sync, NEUROPIXELS_DATA_RELATIVE_PATH) / self.folder
    
    @property
    def qc_path(self) -> pathlib.Path:
        "Expected default path, or alternative if one exists - see `qc_paths` for all available"
        return self.qc_paths[0] if self.qc_paths else QC_PATHS[0] / self.folder
    
    @functools.cached_property
    def qc_paths(self) -> list[pathlib.Path]:
        "Any QC folders that exist"
        return [path / self.folder for path in QC_PATHS if (path / self.folder).exists()]    
    
    @property
    def project(self) -> Project | None:
        if not hasattr(self, "_project"):
            lims_project_name = self.lims.get('project', {}).get('code', '')
            if lims_project_name:
                self._project = Project(lims_project_name)
            else:
                self._project = None
        return self._project
    
    @functools.cached_property
    def lims_data_getter(self) -> dg.data_getter | None:
        try:
            return dg.lims_data_getter(self.id)
        except ConnectionError:
            logger.debug("Connection to lims failed", exc_info=True)
            return None
        except:
            raise

    @property
    def data_dict(self) -> dict:
        if not hasattr(self, "_data_dict"):
            data_getter = self.lims_data_getter
            if not data_getter:
                self._data_dict = {}
            else:
                self._data_dict_orig = data_getter.data_dict  # str paths
                self._data_dict = data_getter.data_dict_pathlib  # pathlib paths
        return self._data_dict

    @property
    def mtrain(self) -> mtrain.MTrain | dict:
        """Info from MTrain on the last behavior session for the mouse on the experiment day"""
        if not hasattr(self, "_mtrain"):
            if not is_connected("mtrain"):
                return {}
            try:
                _ = self.mouse.mtrain
            except mtrain.MouseNotInMTrainError:
                self._mtrain = {}
            except:
                raise
            else:
                self._mtrain = self.mouse.mtrain.last_behavior_session_on(self.date)
        return self._mtrain

    @property
    def foraging_id(self) -> str | None:
        """Foraging ID from MTrain (if an MTrain session is found)."""
        return self.mtrain.get("id", None)

def generate_session(
    mouse: str | int | Mouse,
    user: str | User,
    session_type: Literal["ecephys", "hab"] = "ecephys",
) -> Session:
    if not isinstance(mouse, Mouse):
        mouse = Mouse(mouse)
    if not isinstance(user, User):
        user = User(user)
    if session_type == "ecephys":
        lims_session = lims.generate_ecephys_session(mouse=mouse.lims, user=user.lims)
    elif session_type == "hab":
        lims_session = lims.generate_hab_session(mouse=mouse.lims, user=user.lims)
    session = Session(lims_session)
    # assign instances with data already fetched from lims:
    session._mouse = mouse  
    session._user = user
    return session

def sessions(
    path = NPEXP_ROOT,
    project: str | Projects = None,
    session_type: Literal["ecephys", "behavior"] = "ecephys",
) -> Generator[Session, None, None]:
    """Recursively find Session folders in a directory.

    Project is the common-name among the neuropixels team: 'DR', 'GLO', 'VAR', 'ILLUSION'
    (use the Project enum if unsure)
    """

    if isinstance(project, str):
        project = getattr(Projects, project)

    for path in NPEXP_PATH.iterdir():

        if not path.is_dir():
            continue
        try:
            session = Session(path)
        except (SessionError, FilepathIsDirError):
            continue

        if (
            session_type == "ecephys" and session.is_ecephys_session == False
        ):  # None = unsure and is included
            continue
        if (
            session_type == "behavior" and session.is_ecephys_session
        ):  # None = unsure and is included
            continue

        if project and session.project not in project.value:
            continue

        yield session


if __name__ == "__main__":

    if is_connected("lims2"):
        doctest.testmod()
        # optionflags=(doctest.ELLIPSIS, doctest.NORMALIZE_WHITESPACE, doctest.IGNORE_EXCEPTION_DETAIL)
    else:
        print("LIMS not connected - skipping doctests")



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  9 14:33:49 2020

@author: svc_ccg
"""
from __future__ import annotations

import datetime
import glob
import json
import os
import pathlib

import psycopg2
import psycopg2.extras

class NoBehaviorSessionError(Exception):
    pass


class MultipleBehaviorSessionsError(Exception):
    pass


def get_foraging_id_from_behavior_session(
    mouse_id: int | str, start: datetime.datetime, end: datetime.datetime
) -> str:
    fmt = "%Y-%m-%d %H:%M"
    query = f"""
            SELECT foraging_id
            FROM behavior_sessions bs
                JOIN specimens sp ON sp.donor_id = bs.donor_id
            WHERE date_of_acquisition between '{start.strftime(fmt)}' and '{end.strftime(fmt)}'
            and external_specimen_name = '{mouse_id}'
            """
    cur = get_psql_cursor(get_cred_location())
    cur.execute(query)
    info_list = []
    if cur.rowcount == 0:
        raise NoBehaviorSessionError(
            f"No behavior session found for MID {mouse_id} between {start} and {end}"
        )
    elif cur.rowcount != 0:
        info_list = cur.fetchall()
        if len(info_list) > 1:
            raise MultipleBehaviorSessionsError(
                f"Multiple behavior sessions found for MID {mouse_id} between {start} and {end}"
            )
        elif info_list == []:
            raise NoBehaviorSessionError(
                f"No behavior session found for MID {mouse_id} between {start} and {end}"
            )
        elif len(info_list) == 1 and isinstance(info_list[0], tuple):
            foraging_id = info_list[0][0]
            return foraging_id
        else:
            raise Exception(
                f"Unexpected behavior session info for MID {mouse_id} between {start} and {end}: {info_list}"
            )


# functions from Ahad ------------------------------------------------------------------ #


def get_sess_probes(session_id):
    """Gets a specific session's probes and workflow states
    Parameters
    ----------
    session_id: int
    The sessions's id value
    Returns
    -------
    info_list: str
    A list of the session's passing probes
    """
    EPHYS_PROBE_QRY = """
    SELECT es.workflow_state,
        ARRAY_AGG(ep.id ORDER BY ep.id) AS ephys_probe_ids
    FROM ecephys_sessions es
        LEFT JOIN ecephys_probes ep ON ep.ecephys_session_id = es.id
    WHERE es.id = {}
    GROUP BY es.id
    """
    PROBE_ID_QRY = """
    SELECT ep.name,
        ep.workflow_state,
        ep.storage_directory
    FROM ecephys_probes ep
        JOIN ecephys_sessions es ON es.id = ep.ecephys_session_id
    WHERE ep.id = {}
    """
    cur = get_psql_cursor(get_cred_location())
    lims_query = EPHYS_PROBE_QRY.format(session_id)
    cur.execute(lims_query)
    info_list = []
    if cur.rowcount == 0:
        raise Exception("No data was found for ID {}".format(session_id))
    elif cur.rowcount != 0:
        info_list = cur.fetchall()
        probes_list = []
        probes_id_list = info_list[0][1]
        return probes_id_list
        # returning probe IDs early - need to know only if they exist
        for probe_id in probes_id_list:
            print(probe_id)
            probe_query = PROBE_ID_QRY.format(probe_id)
            cur.execute(probe_query)
            if cur.rowcount == 0:
                raise Exception("No data was found for ID {}".format(session_id))
            else:
                probe_name_status = cur.fetchall()
                probe_status = probe_name_status[0][1]
                probe_name = probe_name_status[0][0]
                probe_storage = probe_name_status[0][2]
                print(probe_storage)
                if (
                    probe_status == "passed" or probe_status == "created"
                ) and probe_storage is not None:
                    if (probe_status) == "passed":
                        print(probe_name)
                    else:
                        print(probe_name + " is created, but not passed")
                    probes_list.append(probe_name)
        return probes_list


def get_cred_location():
    """Gets content of firebase credential file
    Files are ignored and not committed to the repository
    Parameters
    ----------
    Returns
    -------
    cred_json: str
    path to json file storing credential information
    """
    dir = os.path.dirname(__file__)
    cred_json = os.path.join(dir, ".cred", "post_gres.json")
    return cred_json


def get_psql_cursor():
    """Initializes a connection to the postgres database
    Parameters
    ----------
    cred_json: str
    A path to the credential json, which stores the following info:
    dbname: str
    The database name
    user: str
    The username
    host: str
    Host location of the database
    password: str
    The password for the database
    post: int
    The port to connect to
    Returns
    -------
    con: connect
    A connection to the postgres database
    """

    dbname = "lims2"
    user = "limsreader"
    host = "limsdb2"
    password = "limsro"
    port = 5432
    con = psycopg2.connect(
        dbname=dbname, user=user, host=host, password=password, port=port
    )
    con.set_session(readonly=True, autocommit=True)
    return con.cursor()


# functions from Corbett --------------------------------------------------------------- #
class data_getter:
    """parent class for data getter, should be able to
    1) connect to data source
    2) grab experiment data
    3) grab probe data
    """

    def __init__(self, exp_id=None, base_dir=None, cortical_sort=False):

        self.data_dict = {}
        self.cortical_sort = cortical_sort
        self.connect(exp_id, base_dir)
        self.get_exp_data()
        self.get_probe_data()
        self.get_image_data()
        self.data_dict_pathlib = convert_path_str_to_pathlib(self.data_dict)

    def connect(self):
        pass

    def get_exp_data(self):
        pass

    def get_probe_data(self):
        pass

    def get_image_data(self):
        pass


class lims_data_getter(data_getter):
    def connect(self, exp_id, base_dir):

        # set up connection to lims
        self.con = psycopg2.connect(
            dbname="lims2",
            user="limsreader",
            host="limsdb2",
            password="limsro",
            port=5432,
        )
        self.con.set_session(
            readonly=True,
            autocommit=True,
        )
        self.cursor = self.con.cursor(
            cursor_factory=psycopg2.extras.RealDictCursor,
        )

        self.lims_id = exp_id

    def get_exp_data(self):
        """Get all the experiment files
        eg sync, pkls, videos etc
        """
        WKF_QRY = """
            SELECT es.id AS es_id, 
                es.name AS es,
                es.storage_directory,
                es.workflow_state,
                es.date_of_acquisition,
                es.stimulus_name,
                es.foraging_id as foraging_id,
                sp.external_specimen_name,
                isi.id AS isi_experiment_id,
                e.name AS rig,
                u.login AS operator,
                p.code AS project,
                wkft.name AS wkft, 
                wkf.storage_directory || wkf.filename AS wkf_path,
                bs.storage_directory AS behavior_dir
            FROM ecephys_sessions es
                JOIN specimens sp ON sp.id = es.specimen_id
                LEFT JOIN isi_experiments isi ON isi.id = es.isi_experiment_id
                LEFT JOIN equipment e ON e.id = es.equipment_id
                LEFT JOIN users u ON u.id = es.operator_id
                JOIN projects p ON p.id = es.project_id
                LEFT JOIN well_known_files wkf ON wkf.attachable_id = es.id
                LEFT JOIN well_known_file_types wkft ON wkft.id=wkf.well_known_file_type_id
                LEFT JOIN behavior_sessions bs ON bs.foraging_id = es.foraging_id
            WHERE es.id = {}
            ORDER BY es.id
            """
        #

        self.cursor.execute(WKF_QRY.format(self.lims_id))
        exp_data = self.cursor.fetchall()
        if not exp_data:
            return
        self.data_dict.update(
            exp_data[0]
        )  # update data_dict to have all the experiment metadata
        [
            self.data_dict.pop(key) for key in ["wkft", "wkf_path"]
        ]  # ...but remove the wkf stuff

        for e in exp_data:
            wkft = e["wkft"]
            wkf_path = e["wkf_path"]
            self.data_dict[wkft] = convert_lims_path(wkf_path)

        self.translate_wkf_names()

        behavior_dir = convert_lims_path(self.data_dict["behavior_dir"])
        self.data_dict["behavior_pkl"] = glob_file(os.path.join(behavior_dir, "*.pkl"))
        if self.data_dict["date_of_acquisition"] is not None:
            self.data_dict["datestring"] = self.data_dict[
                "date_of_acquisition"
            ].strftime("%Y%m%d")
        else:
            self.data_dict["datestring"] = ""
        self.data_dict["es_id"] = str(self.data_dict["es_id"])

    def get_image_data(self):
        """Get all the images associated with this experiment"""

        IMAGE_QRY = """
            SELECT es.id AS es_id, es.name AS es, imt.name AS image_type, es.storage_directory || im.jp2 AS image_path
            FROM ecephys_sessions es
                JOIN observatory_associated_data oad ON oad.observatory_record_id = es.id AND oad.observatory_record_type = 'EcephysSession'
                JOIN images im ON im.id=oad.observatory_file_id AND oad.observatory_file_type = 'Image'
                JOIN image_types imt ON imt.id=im.image_type_id
            WHERE es.id = {}
            ORDER BY es.id, imt.name;
            """

        self.cursor.execute(IMAGE_QRY.format(self.lims_id))
        image_data = self.cursor.fetchall()

        # FOR NOW JUST ASSUME IMAGES ARE IN THE D1 UPLOAD DIRECTORY
        # get D1 directory (assume this is where the sync file is)

        for im in image_data:
            name = im["image_type"]
            path = convert_lims_path(im["image_path"])
            # self.data_dict[name] = convert_lims_path(path)
            self.data_dict[name] = path

    def get_probe_data(self):
        """Get sorted ephys data for each probe

        TODO: make this actually use the well known file types,
        rather than just grabbing the base directories

        """

        WKF_PROBE_QRY = """
            SELECT es.id AS es_id, 
                es.name AS es, 
                ep.name AS ep, 
                ep.id AS ep_id, 
                wkft.name AS wkft, 
                wkf.storage_directory || wkf.filename AS wkf_path
            FROM ecephys_sessions es
                JOIN ecephys_probes ep ON ep.ecephys_session_id=es.id
                LEFT JOIN well_known_files wkf ON wkf.attachable_id = ep.id
                LEFT JOIN well_known_file_types wkft ON wkft.id=wkf.well_known_file_type_id
            WHERE es.id = {} 
            ORDER BY es.id, ep.name;
            """
        self.cursor.execute(WKF_PROBE_QRY.format(self.lims_id))
        probe_data = self.cursor.fetchall()

        p_info = [p for p in probe_data if p["wkft"] == "EcephysSortedAmplitudes"]

        getnesteddir = lambda x: os.path.dirname(os.path.dirname(os.path.dirname(x)))
        probe_bases = [convert_lims_path(getnesteddir(pi["wkf_path"])) for pi in p_info]
        # probe_bases = [convert_lims_path(os.path.dirname(pi['wkf_path'])) for pi in p_info]

        self.data_dict["data_probes"] = []
        for pb in probe_bases:
            probeID = pb[-1]
            self.data_dict["data_probes"].append(probeID)
            self.data_dict["probe" + probeID] = pb
            self.data_dict["lfp" + probeID] = pb
            info_json = glob_file(os.path.join(pb, "*probe_info*json"))
            self.data_dict["probe" + probeID + "_info"] = info_json

        raw = [p for p in probe_data if p["wkft"] == "EcephysProbeRawData"]
        name_suffix = {
            "probeA": "ABC",
            "probeB": "ABC",
            "probeC": "ABC",
            "probeD": "DEF",
            "probeE": "DEF",
            "probeF": "DEF",
        }
        for r in raw:
            probeID = r["ep"]
            name = r["wkft"] + name_suffix[probeID]
            path = convert_lims_path(r["wkf_path"])

            if (
                not name + "_settings" in self.data_dict
                or self.data_dict[name + "_settings"] is None
            ):
                self.data_dict[name + "_settings"] = path

            npx2_path = glob_file(os.path.join(os.path.dirname(path), "*npx2"))
            if not name in self.data_dict or self.data_dict[name] is None:
                self.data_dict[name] = npx2_path

        self.probe_data = probe_data

    def storage_directory(self) -> str | None:
        """Get the storage directory for this experiment"""
        if self.data_dict:
            return "/" + self.data_dict["storage_directory"]
        WKF_QRY = """
            SELECT es.storage_directory
            FROM ecephys_sessions es
            WHERE es.id = {}
            """
        self.cursor.execute(WKF_QRY.format(self.lims_id))
        exp_data = self.cursor.fetchall()
        if exp_data and exp_data[0]["storage_directory"]:
            return "/" + exp_data[0]["storage_directory"]
        return None

    def translate_wkf_names(self):
        wkf_dict = {
            "MappingPickle": "mapping_pkl",
            "EcephysReplayStimulus": "replay_pkl",
            "EcephysRigSync": "sync_file",
            "OptoPickle": "opto_pkl",
        }

        for wkf in wkf_dict:
            if wkf in self.data_dict:
                self.data_dict[wkf_dict[wkf]] = self.data_dict[wkf]


class local_data_getter(data_getter):
    def connect(self, exp_id, base_dir):

        if os.path.exists(base_dir):
            self.base_dir = base_dir
        else:
            print("Invalid base directory: " + base_dir)

    def get_exp_data(self):
        file_glob_dict = {
            "mapping_pkl": ["*mapping*.pkl", "*stim.pkl"],
            "replay_pkl": "*replay*.pkl",
            "behavior_pkl": "*behavior*.pkl",
            "opto_pkl": "*opto*.pkl",
            "sync_file": "*.sync",
            "RawEyeTrackingVideo": ["*.eye.avi", "*eye.mp4"],
            "RawBehaviorTrackingVideo": ["*behavior.avi", "*behavior.mp4"],
            "RawFaceTrackingVideo": ["*face.avi", "*face.mp4"],
            "RawEyeTrackingVideoMetadata": "*eye.json",
            "RawBehaviorTrackingVideoMetadata": "*behavior.json",
            "RawFaceTrackingVideoMetadata": "*face.json",
            "EcephysPlatformFile": "*platformD1.json",
            "NewstepConfiguration": "*motor-locs.csv",
        }

        for fn in file_glob_dict:
            if isinstance(file_glob_dict[fn], list):
                paths = [
                    glob_file(os.path.join(self.base_dir, f))
                    for f in file_glob_dict[fn]
                ]
                path = [p for p in paths if not p is None]
                if len(path) > 0:
                    self.data_dict[fn] = path[0]
            else:
                filepath = glob_file(os.path.join(self.base_dir, file_glob_dict[fn]))
                if filepath is not None:
                    self.data_dict[fn] = filepath

        basename = os.path.basename(self.base_dir)
        self.data_dict["es_id"] = basename.split("_")[0]
        self.data_dict["external_specimen_name"] = basename.split("_")[1]
        self.data_dict["datestring"] = basename.split("_")[2]
        self.data_dict["rig"] = self.get_rig_from_platform()

    def get_platform_info(self):

        platform_file = self.data_dict["EcephysPlatformFile"]
        with open(platform_file, "r") as file:
            self.platform_info = json.load(file)

    def get_rig_from_platform(self):

        if not hasattr(self, "platform_info"):
            self.get_platform_info()

        return self.platform_info["rig_id"]

    def get_probe_data(self):
        self.data_dict["data_probes"] = []

        # get probe dirs
        for probeID in "ABCDEF":
            if self.cortical_sort:
                probe_base = glob_file(
                    os.path.join(self.base_dir, "cortical*probe" + probeID + "_sorted")
                )
                lfp_base = glob_file(
                    os.path.join(self.base_dir, "*probe" + probeID + "_sorted")
                )
            else:
                probe_base = glob_file(
                    os.path.join(self.base_dir, "*probe" + probeID + "_sorted")
                )
                lfp_base = probe_base

            if probe_base is not None:
                self.data_dict["data_probes"].append(probeID)
                self.data_dict["probe" + probeID] = probe_base

                metrics_file = glob_file(
                    os.path.join(
                        probe_base, r"continuous\Neuropix-PXI-100.0\metrics.csv"
                    )
                )
                self.data_dict["probe" + probeID + "_metrics"] = metrics_file

                info_json = glob_file(os.path.join(probe_base, "*probe_info*json"))
                self.data_dict["probe" + probeID + "_info"] = info_json

                channel_map = glob_file(
                    os.path.join(
                        probe_base, r"continuous\Neuropix-PXI-100.0\channel_map.npy"
                    )
                )
                self.data_dict["probe" + probeID + "_channel_map"] = channel_map

            if lfp_base is not None:
                self.data_dict["lfp" + probeID] = lfp_base

    def get_image_data(self):

        # GET PROBE DEPTH IMAGES
        for probeID in self.data_dict["data_probes"]:

            probe_base = self.data_dict["probe" + probeID]
            probe_depth_image = glob_file(os.path.join(probe_base, "probe_depth*.png"))
            if probe_depth_image is not None:
                self.data_dict["probe_depth_" + probeID] = probe_depth_image

        # GET OTHER IMAGE FILES
        # image_files = [k for k in D1_local if 'image' in k]
        image_files = [
            "EcephysPostExperimentLeft",
            "EcephysPostExperimentRight",
            "EcephysPostInsertionLeft",
            "EcephysPostInsertionRight",
            "EcephysPostStimulusLeft",
            "EcephysPostStimulusRight",
            "EcephysPreExperimentLeft",
            "EcephysPreExperimentRight",
            "EcephysPreInsertionLeft",
            "EcephysPreInsertionRight",
            "EcephysInsertionLocationImage",
            "EcephysOverlayImage",
            "EcephysBrainSurfaceLeft",
            "EcephysBrainSurfaceRight",
        ]

        for im in image_files:
            im_info = D1_local[im]
            im_file = glob_file(os.path.join(self.base_dir, im_info["rel_path"]))

            self.data_dict[im] = im_file


def glob_file(file_path):
    f = glob.glob(file_path)
    if len(f) > 0:
        return f[0]
    else:
        return None


def convert_lims_path(path):
    if path is not None:
        new_path = r"\\" + os.path.normpath(path)[1:]
    else:
        new_path = ""

    return new_path


def convert_path_str_to_pathlib(data_dict_orig) -> dict:
    """
    >>> orig = {0: '\\\\allen\\programs\\mindscope'}
    >>> test = convert_path_str_to_pathlib(orig)
    >>> test[0].as_posix()
    '//allen/programs/mindscope'
    
    >>> test[0] != orig[0]
    True
    
    >>> test = convert_path_str_to_pathlib({1: '/allen/programs/mindscope'})
    >>> test[1].as_posix()
    '//allen/programs/mindscope'
    """
    data_dict = data_dict_orig.copy()
    for k, v in data_dict.items():
        if isinstance(v, str) and (v.startswith("/") or v.startswith("\\")):
            v = v.replace("\\", "/")
            if v[:2] != "//":
                v = "/" + v
            data_dict[k] = pathlib.Path(v)
    return data_dict

if __name__ == "__main__":
    import doctest
    doctest.testmod()


"""
Critical components of an NP experiment workflow that come from lims2.

- user/operator info
- mouse info 
    - isi experiment
    - project name

Combined, these are sufficient to create a new ecephys session entry in lims2.

Tools are also provided to construct commonly used abbreviations or folder names, and to
reverse these 'NP' formats to reconstruct the Info objects used during an
experiment workflow.

"""
from __future__ import annotations

import abc
import collections
import datetime
import doctest
import functools
import json
import pathlib
import re
from typing import Any, Callable, Type, Union

import np_logging
import requests

logger = np_logging.getLogger(__name__)

def requester(url: str, *args) -> dict:
    request = url.format(*args)  # .replace(";", "%3B")
    logger.debug(f"Requesting {request}")
    response = requests.get(request)
    if response.status_code != 200:
        raise ValueError(f"Bad response from {request}: {response.status_code}")
    return response.json()


request = lambda url: functools.partial(requester, url)
donor_info = request("http://lims2/donors/info/details.json?external_donor_name={}")
user_info = request("http://lims2/users.json?login={}")
ecephys_info = request("http://lims2/ecephys_sessions.json?id={}")
behavior_info = request("http://lims2/behavior_sessions.json?id={}")
isi_info = request("http://lims2/specimens/isi_experiment_details/{}.json")
project_info = request("http://lims2/projects.json?code={}")


class LIMS2InfoBaseClass(collections.UserDict, abc.ABC):
    "Store details for an object in a dict-like. The commonly-used format of its name, e.g. '366122' for a mouse ID, can be obtained by converting to str()."

    np_id: int | str
    "Commonly-used format of the object's value among the neuropixels team e.g. for a mouse -> the labtracks ID (366122)."

    _type: Type = NotImplemented
    _get_info: Callable[[str | int], dict] = NotImplemented

    def __init__(self, np_id: str | int):
        self.np_id = self.__class__._type(np_id)
        super().__init__()

    def __getitem__(self, key):
        self.fetch()
        return super().__getitem__(key)

    def __str__(self):
        return str(self.np_id)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.np_id!r})"

    def __bool__(self):
        try:
            self.info_from_lims()
        except ValueError:
            return False
        else:
            return True

    def info_from_lims(self) -> dict[str, Any]:
        "Return the object's info from lims database or raises a ValueError if not found."
        try:
            return self._get_info(self.np_id)[0]
        except IndexError:
            raise ValueError(
                f"Could not find {self.__class__.__name__} {self.np_id} in lims"
            ) from None

    def fetch(self):
        "Fetch the object's info from lims once."
        if not self.data:
            try:
                info = self.info_from_lims()
            except ValueError:
                self.data = {}
            else:
                self.data = dict(**info)

    def keys(self):
        "Fetch before returning keys."
        self.fetch()
        return self.data.keys()

    def items(self):
        "Fetch before returning keys."
        self.fetch()
        return self.data.items()

    def values(self):
        "Fetch before returning keys."
        self.fetch()
        return self.data.values()

    @abc.abstractproperty
    def lims_id(self):
        "LIMS2 ID for the object, usually different to the np_id."
        return NotImplemented
    
    @property
    def id(self) -> str | int:
        return self.lims_id
    
    # end of baseclass properties & methods ------------------------------ #


class LIMS2MouseInfo(LIMS2InfoBaseClass):
    """
    Mouse info from lims stored in a dict, with a string method for the commonly-used format of its name.

    >>> mouse = LIMS2MouseInfo(366122)

    >>> str(mouse)
    '366122'

    >>> mouse['id']
    657428267

    >>> mouse.project_name
    'NeuropixelPlatformDevelopment'
    """

    _type = int
    _get_info = donor_info

    @property
    def lims_id(self) -> int:
        return self["specimens"][0]["id"]

    @property
    def isi_info(self) -> dict | None:
        "Info from lims about the mouse's ISI experiments."
        if not hasattr(self, "_isi_info"):
            response = isi_info(str(self.np_id))
            if response:
                self._isi_info = response[0]
            else:
                self._isi_info = None
        return self._isi_info

    @property
    def isi_id(self) -> int | None:
        "ID of the mouse's most recent ISI experiment not marked `failed`."
        exps: list = self.isi_info["isi_experiments"]
        exps.sort(key=lambda x: x["id"], reverse=True)
        for exp in exps:
            if exp["workflow_state"] != "failed":
                return exp["id"]
        return None

    @property
    def project_id(self) -> int:
        "ID of the the project the mouse belongs to."
        return self["specimens"][0]["project_id"]

    @property
    def project_name(self) -> str:
        "PascalCase name of the project the mouse belongs to."
        return self["specimens"][0]["project"]["code"]

    @property
    def path(self) -> pathlib.Path:
        "Allen network dir where the mouse's sessions are stored."
        return pathlib.Path(self["specimens"][0]["storage_directory"])


class LIMS2UserInfo(LIMS2InfoBaseClass):
    "Store details for a user/operator."

    _type = str
    _get_info = user_info

    @property
    def lims_id(self) -> int:
        return self["id"]


class LIMS2ProjectInfo(LIMS2InfoBaseClass):
    "Store details for a project."

    _type = str
    _get_info = project_info

    @property
    def lims_id(self) -> int:
        return self["id"]


class LIMS2SessionInfo(LIMS2InfoBaseClass):
    "Store details for an ecephys or behavior session."

    _type = int
    _get_info = ecephys_info  # default - behavior_info may be used instead

    def __str__(self):
        return f"{self.np_id}"

    @property
    def lims_id(self) -> int:
        return self.np_id

    @property
    def session(self) -> str:
        self.cast()
        return self._session

    @property
    def folder(self):
        self.cast()
        return self.get_folder()

    def info_from_lims(self) -> dict:
        try:
            data = super().info_from_lims()  # with ecephys_info
        except ValueError:
            response = behavior_info(self.np_id)  # try behavior_info instead
            if not response:
                raise
            self._session = "behavior"
            data = response[0]
        else:
            self._session = "ecephys"
        self.cast()
        return data

    def cast(self):
        if not hasattr(self, "_session"):
            _ = self.fetch()  # trigger fetching data from lims
        self.__class__ = (
            LIMS2EcephysSessionInfo if self._session == "ecephys" else LIMS2BehaviorSessionInfo
        )


class LIMS2EcephysSessionInfo(LIMS2SessionInfo):
    "Don't instantiate directly, use LIMS2SessionInfo and class will be converted after info fetched from lims."

    def get_folder(self) -> str:
        return f"{self.np_id}_{self['specimen']['external_specimen_name']}_{self['name'].replace('HAB_', '')[:8]}"

    def __repr__(self):
        return f"{__class__.__bases__[0].__name__}({self.np_id!r})"


class LIMS2BehaviorSessionInfo(LIMS2SessionInfo):
    "Don't instantiate directly, use LIMS2SessionInfo and class will be converted after info fetched from lims."

    def get_folder(self) -> str:
        return f"{self.np_id}_{self['donor']['name'].split('-')[-1]}_{self['ecephys_session']['name'][:8]}"

    def __repr__(self):
        return f"{__class__.__bases__[0].__name__}({self.np_id!r})"


# end of classes ----------------------------------------------------------------------- #


def generate_ecephys_session(
    mouse: str | int | LIMS2MouseInfo,
    user: str | LIMS2UserInfo,
) -> LIMS2SessionInfo:
    "Create a new session and return an object instance with its info."

    if not isinstance(mouse, LIMS2MouseInfo):
        mouse = LIMS2MouseInfo(mouse)
    if not isinstance(user, LIMS2UserInfo):
        user = LIMS2UserInfo(user)

    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    request_json = {
        "specimen_id": mouse.lims_id,
        "project_id": mouse.project_id,
        "isi_experiment_id": mouse.isi_id,
        "name": f"{timestamp}_{user.lims_id}",
        "operator_id": user.lims_id,
    }
    url = "http://lims2/observatory/ecephys_session/create"
    response = requests.post(url, json=request_json)
    decoded_dict = json.loads(response.content.decode("utf-8"))
    new_session_id = decoded_dict["id"]
    if not new_session_id:
        raise ValueError(f"Failed to create session: {decoded_dict}")
    return LIMS2SessionInfo(new_session_id)

def generate_hab_session(
    mouse: str | int | LIMS2MouseInfo,
    user: str | LIMS2UserInfo,
) -> LIMS2SessionInfo:
    "Create a new session and return an object instance with its info."

    if not isinstance(mouse, LIMS2MouseInfo):
        mouse = LIMS2MouseInfo(mouse)
    if not isinstance(user, LIMS2UserInfo):
        user = LIMS2UserInfo(user)

    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    request_json = {
        "specimen_id": mouse.lims_id,
        "project_id": mouse.project_id,
        "isi_experiment_id": mouse.isi_id,
        "name": f"HAB_{timestamp}_{user.lims_id}",
        "operator_id": user.lims_id,
    }
    url = "http://lims2/observatory/ecephys_session/create"
    response = requests.post(url, json=request_json)
    decoded_dict = json.loads(response.content.decode("utf-8"))
    new_session_id = decoded_dict["id"]
    if not new_session_id:
        raise ValueError(f"Failed to create session: {decoded_dict}")
    return LIMS2SessionInfo(new_session_id)


def find_session_folder_string(path: Union[str, pathlib.Path]) -> str | None:
    """Extract [8+digit session ID]_[6-digit mouse ID]_[6-digit date
    str] from a file or folder path"""
    session_reg_exp = r"[0-9]{8,}_[0-9]{6}_[0-9]{8}"
    session_folders = re.findall(session_reg_exp, str(path))
    if session_folders:
        if not all(s == session_folders[0] for s in session_folders):
            raise ValueError(
                f"Mismatch between session folder strings - file may be in the wrong folder: {path}"
            )
        return session_folders[0]
    return None


def info_classes_from_session_folder(
    session_folder: str,
) -> tuple[LIMS2SessionInfo, LIMS2MouseInfo, LIMS2UserInfo]:
    "Reconstruct Info objects from a session folder string."
    folder = find_session_folder_string(session_folder)
    if not folder:
        raise ValueError(f"{session_folder} is not a valid session folder")

    session = LIMS2SessionInfo(folder.split("_")[0])
    mouse = LIMS2MouseInfo(folder.split("_")[1])
    user = LIMS2UserInfo(session["operator"]["login"])

    return (session, mouse, user)


if __name__ == "__main__":
    doctest.testmod()


from __future__ import annotations

import datetime
import doctest
import enum
import functools
import json
import os
import pathlib
import re
import subprocess
import sys
from typing import Literal, Union

import np_logging
import requests

import np_session.components.paths as paths
from np_session.databases.data_getters import lims_data_getter
from np_session.databases.lims2 import LIMS2SessionInfo

logger = np_logging.getLogger(__name__)

PathLike = Union[str, bytes, os.PathLike, pathlib.Path]
# https://peps.python.org/pep-0519/#provide-specific-type-hinting-support
# PathLike inputs are converted to pathlib.Path objects for os-agnostic filesystem operations
# os.fsdecode(path: PathLike) is used where only a string is required

RE_SESSION_ID = re.compile("[0-9]{8,}")
RE_FOLDER = re.compile("[0-9]{8,}_[0-9]{6}_[0-9]{8}")
RE_PROBES = re.compile("(?<=_probe)_?(([A-F]+)|([0-5]))")

REPLACED_COMP_ID: dict[str, tuple[datetime.date, str]] = {
    'NP.0-Acq': (datetime.date(2022, 10, 18), "W10DT05515"),
    'NP.1-Acq': (datetime.date(2022, 10, 27), "W10DT05501"),
    'NP.2-Acq': (datetime.date(2022, 7, 14), "W10DT05517"),
    'NP.0-Stim': (datetime.date(2023, 2, 7), "W10DT713938"),
    'NP.1-Stim': (datetime.date(2023, 1, 19), "W10DT713942"),
}

def old_hostname(comp_id: str, date: datetime.date) -> str | None:
    """Return the hostname for a computer that was replaced, if `date` predates the switchover.
    
    For a date before the hostname changed:
    >>> old_hostname('NP.1-Acq', datetime.date(2022, 6, 18))
    'W10DT05501'

    For a date after the hostname changed, nothing is returned: 
    >>> old_hostname('NP.1-Acq', datetime.date(2023, 6, 18))

    """
    if not (replaced := REPLACED_COMP_ID.get(comp_id)):
        return
    if date < replaced[0]:
        return replaced[1]

class Host(enum.Enum):
    LIMS = "lims2"
    MTRAIN = "mtrain"
    MINDSCOPE_SERVER = ZK = "eng-mindscope"
    MPE_SERVER = "aibspi"


def is_connected(host: str | Host) -> bool:
    "Use OS's `ping` cmd to check if `host` is connected."
    command = ["ping", "-n" if "win" in sys.platform else "-c", "1", host]
    try:
        return subprocess.call(command, stdout=subprocess.PIPE, timeout=0.1) == 0
    except subprocess.TimeoutExpired:
        return False


def is_lims_path(path: PathLike) -> bool:
    """
    >>> is_lims_path('//allen/programs/mindscope/production/dynamicrouting/prod0/specimen_1200280254/ecephys_session_1234028213')
    True
    """
    parts = pathlib.Path(path).parts
    return "production" in parts and "incoming" not in parts


def is_valid_session_id(session_id: int | str) -> bool:
    """
    >>> is_valid_session_id('1234028213')
    True
    >>> is_valid_session_id('abcdefg')
    False
    """
    try:
        _ = int(session_id)
    except ValueError:
        return False
    return bool(LIMS2SessionInfo(session_id))


def lims_session_id(path: PathLike) -> str | None:
    """
    Get valid lims ecephys/behavior session id from str or path.

    >>> lims_session_id('//allen/programs/mindscope/production/dynamicrouting/prod0/specimen_1200280254/ecephys_session_1234028213')
    '1234028213'
    """

    path = os.fsdecode(path)
    if is_lims_path(path):
        from_lims_path = re.search("(?<=_session_)\\d+", os.fsdecode(path))
        if from_lims_path:
            return from_lims_path.group(0)
    for i in RE_SESSION_ID.findall(path):
        if is_valid_session_id(i):
            return i


def folder(path: PathLike) -> str | None:
    """
    Extract [8+digit lims session ID]_[6-digit labtracks mouse ID]_[6-digit datestr] from a str or path.

    >>> folder('//allen/programs/mindscope/workgroups/np-exp/1234028213_640887_20221219/image.png')
    '1234028213_640887_20221219'
    """

    session_folders = RE_FOLDER.findall(os.fsdecode(path))

    if not session_folders:
        return folder_from_lims_id(path)
    if not all(s == session_folders[0] for s in session_folders):
        logger.warning(
            f"Mismatch between session folder strings - file may be in the wrong folder: {path!r}"
        )
    return session_folders[0]


def folder_from_lims_id(path: PathLike) -> str | None:
    """
    Get the session folder string ([lims-id]_[mouse-id]_[date]) from a string or path containing a possible lims id.

    >>> folder_from_lims_id('//allen/programs/mindscope/production/dynamicrouting/prod0/specimen_1200280254/ecephys_session_1234028213')
    '1234028213_640887_20221219'

    >>> folder_from_lims_id('1234028213')
    '1234028213_640887_20221219'
    """
    session_id = lims_session_id(path)
    if session_id is None:
        return None
    lims_data = lims_data_getter(session_id)
    return ("_").join(
        [
            lims_data.lims_id,
            lims_data.data_dict["external_specimen_name"],
            lims_data.data_dict["datestring"],
        ]
    )


def folder_from_lims_id(path: PathLike) -> str | None:
    """
    Get the session folder string ([lims-id]_[mouse-id]_[date]) from a string or path containing a possible lims id.

    >>> folder_from_lims_id('//allen/programs/mindscope/production/dynamicrouting/prod0/specimen_1200280254/ecephys_session_1234028213')
    '1234028213_640887_20221219'

    >>> folder_from_lims_id('1234028213')
    '1234028213_640887_20221219'
    """
    session_id = lims_session_id(path)
    if session_id is None:
        return None
    return LIMS2SessionInfo(session_id).folder


@functools.lru_cache(maxsize=None)
def lims_json_content(lims_id: int | str) -> dict | None:
    if not is_valid_session_id(lims_id):
        raise ValueError(f"{lims_id} is not a valid lims session id")
    if not is_connected("lims2"):
        raise ConnectionError("Could not connect to lims")
    for session_type in ["ecephys_sessions", "behavior_sessions"]:
        response = requests.get(f"http://lims2/{session_type}/{lims_id}.json?")
        if response.status_code == 200:
            return response.json()
    logger.warning(f"Could not find json content for lims session id {lims_id}")
    return None


def is_new_ephys_folder(path: PathLike) -> bool:
    "Contains subfolders with raw data from OpenEphys v0.6.0+ (format switched 2022 on NP.0,1,2)"
    path = pathlib.Path(path)
    if path.match("_probe*"):
        return path.match("Record Node*") or path.rglob("Record Node*")
    return bool(list(path.glob("*_probe*/Record Node*")))


def files_manifest(
        project_name: str,
        session_str: str = '',
        session_type: Literal['D1', 'D2', 'habituation'] = 'D1',
    ) -> dict[str, dict[str,str]]:
    """Return a list of files that could be entered directly into a platform json `files` key.
    - project_name: corresponds to a manifet template 
    - session_str: [lims_id]_[mouse_id]_[session_id], will replace a placeholder in a manifest template
    """
    if session_type not in ['D1', 'habituation', 'D2']:
        raise ValueError(f'{session_type} is not a valid session type')
    
    template = paths.TEMPLATES_ROOT / session_type / f"{project_name}.json"
    
    x = json.loads(template.read_bytes())
    # convert dict to str
    # replace % with session string
    # switch ' and " so we can convert str back to dict with json.loads()
    return json.loads(str(x).replace('%', str(session_str)).replace('\'','"'))


if __name__ == "__main__":

    if is_connected("lims2"):
        doctest.testmod()
        # optionflags=(doctest.ELLIPSIS, doctest.NORMALIZE_WHITESPACE, doctest.IGNORE_EXCEPTION_DETAIL)
    else:
        print("LIMS not connected - skipping doctests")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['np_session',
 'np_session.components',
 'np_session.databases',
 'np_session.jobs']

package_data = \
{'': ['*']}

install_requires = \
['np_logging>=0.3.8',
 'psycopg2-binary>=2,<3',
 'requests>=2,<3',
 'typing-extensions']

setup_kwargs = {
    'name': 'np-session',
    'version': '0.2.0',
    'description': 'Tools for accessing data, metadata, and jobs related to ecephys and behavior sessions for the Mindscope Neuropixels team.',
    'long_description': "# np_session\n\n\n### *For use on internal Allen Institute network*\n\n\n```python\nfrom np_session import Session\n\n# initialize with a lims session ID or a string containing one: \n>>> session = Session('c:/1116941914_surface-image1-left.png') \n>>> session.lims.id\n1116941914\n>>> session.folder\n'1116941914_576323_20210721'\n>>> session.project\n'BrainTV Neuropixels Visual Behavior'\n>>> session.is_ecephys_session\nTrue\n>>> session.rig.acq # hostnames reflect the computers used during the session, not necessarily the current machines\n'W10DT05515'\n\n# some properties are objects with richer information:\n\n# - `pathlib` objects for filesystem paths:\n>>> session.lims_path.as_posix()\n'//allen/programs/braintv/production/visualbehavior/prod0/specimen_1098595957/ecephys_session_1116941914'\n>>> session.data_dict['es_id']\n'1116941914'\n\n# - `datetime` objects for easy date manipulation:\n>>> session.date\ndatetime.date(2021, 7, 21)\n\n# - dictionaries from lims (loaded lazily):\n>>> session.mouse\nMouse(576323)\n>>> session.mouse.lims\nLIMS2MouseInfo(576323)\n>>> session.mouse.lims.id\n1098595957\n>>> session.mouse.lims['full_genotype']\n'wt/wt'\n\n# ...with a useful string representation:\n>>> str(session.mouse)\n'576323'\n>>> str(session.project)\n'NeuropixelVisualBehavior'\n>>> str(session.rig)        # from `np_config` package\n'NP.0'\n```\n",
    'author': 'Ben Hardcastle',
    'author_email': 'ben.hardcastle@alleninstitute.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
